"use strict";
/**
 * MediaWiki Api for Axios
 * Provides the API call methods similar to `mw.Api` at non-mw environments
 *
 * @author Dragon-Fish <dragon-fish@qq.com>
 * @license MIT
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.WikiSaikouError = exports.WikiSaikouErrorCode = exports.ForeignApi = exports.MwApi = exports.MediaWikiForeignApi = exports.MediaWikiApi = void 0;
const reactivity_1 = require("@vue/reactivity");
const fexios_1 = require("fexios");
class MediaWikiApi {
    constructor(baseURL, options) {
        var _a, _b;
        this.cookies = {};
        // For MediaWiki browser environment
        if (!baseURL && typeof window === 'object' && window.mediaWiki) {
            const { wgServer, wgScriptPath } = ((_b = (_a = window.mediaWiki) === null || _a === void 0 ? void 0 : _a.config) === null || _b === void 0 ? void 0 : _b.get(['wgServer', 'wgScriptPath'])) ||
                {};
            if (typeof wgServer === 'string' && typeof wgScriptPath === 'string') {
                baseURL = `${wgServer}${wgScriptPath}/api.php`;
            }
        }
        if (typeof baseURL !== 'string') {
            throw new Error('baseURL is undefined');
        }
        // Init
        this.baseURL = (0, reactivity_1.ref)(baseURL);
        this.tokens = {};
        this.defaultParamsRef = (0, reactivity_1.ref)({});
        this.defaultOptionsRef = (0, reactivity_1.ref)({});
        // Set default values
        this.defaultParams = {
            action: 'query',
            errorformat: 'plaintext',
            format: 'json',
            formatversion: 2,
        };
        this.defaultOptions = options || {};
        this.requestHandlerRef = (0, reactivity_1.computed)(() => {
            const instance = MediaWikiApi.createRequestHandler(this.baseURL.value, {
                ...this.defaultOptions,
                query: this.defaultParams,
            });
            // Handle cookies for Node.js
            if (!('document' in globalThis)) {
                instance.interceptors.request.use((ctx) => {
                    ctx.headers = ctx.headers || {};
                    ctx.headers['cookie'] = Object.keys(this.cookies)
                        .map((name) => `${name}=${this.cookies[name]}`)
                        .join(';');
                    return ctx;
                });
                instance.interceptors.response.use((ctx) => {
                    const cookieHeaders = ctx.rawResponse.headers.get('set-cookie');
                    const rawCookies = cookieHeaders === null || cookieHeaders === void 0 ? void 0 : cookieHeaders.split(',').map((i) => i.trim());
                    rawCookies === null || rawCookies === void 0 ? void 0 : rawCookies.forEach((i) => {
                        const [name, ...value] = i.split(';')[0].split('=');
                        this.cookies[name] = value.join('=');
                    });
                    return ctx;
                });
            }
            return instance;
        });
    }
    static normalizeParamValue(item) {
        if (Array.isArray(item)) {
            return item.join('|');
        }
        else if (typeof item === 'boolean') {
            return item ? '1' : undefined;
        }
        else if (typeof item === 'number') {
            return '' + item;
        }
        else {
            return item;
        }
    }
    static createRequestHandler(baseURL, options = {}) {
        const instance = new fexios_1.Fexios(options);
        instance.baseConfigs.baseURL = baseURL;
        // Adjust body
        instance.on('beforeInit', (ctx) => {
            var _a;
            if (((_a = ctx.method) === null || _a === void 0 ? void 0 : _a.toLowerCase()) !== 'post') {
                return ctx;
            }
            if (typeof ctx.body === 'object' &&
                ctx.body !== null &&
                !(ctx.body instanceof URLSearchParams) &&
                !(ctx.body instanceof FormData)) {
                const body = ctx.body;
                Object.keys(body).forEach((key) => {
                    const data = MediaWikiApi.normalizeParamValue(body[key]);
                    if (typeof data === 'undefined' || data === null) {
                        delete body[key];
                    }
                    else if (data !== body[key]) {
                        body[key] = data;
                    }
                });
                ctx.body = new URLSearchParams(ctx.body);
            }
            if ((globalThis.FormData && ctx.body instanceof FormData) ||
                ctx.body instanceof URLSearchParams) {
                const body = ctx.body;
                // Adjust params
                body.forEach((value, key) => {
                    const data = MediaWikiApi.normalizeParamValue(value);
                    if (typeof data === 'undefined' || data === null) {
                        body.delete(key);
                    }
                    else if (data !== value) {
                        body.set(key, data);
                    }
                });
                // Adjust query
                const searchParams = new URLSearchParams(ctx.query);
                !searchParams.has('format') &&
                    searchParams.set('format', '' + (body.get('format') || 'json'));
                !searchParams.has('formatversion') &&
                    searchParams.set('formatversion', '' + (body.get('formatversion') || '2'));
                body.has('origin') &&
                    searchParams.set('origin', '' + body.get('origin'));
                ctx.query = Object.fromEntries(searchParams);
                // DONT REMOVE THIS
                // TODO: Remove duplicate parameters. There should be a better solution.
                body.has('action') && (ctx.query.action = '' + body.get('action'));
            }
            return ctx;
        });
        // Adjust query
        instance.on('beforeInit', (ctx) => {
            ctx.query = ctx.query;
            for (const key in ctx.query) {
                const data = MediaWikiApi.normalizeParamValue(ctx.query[key]);
                if (typeof data === 'undefined' || data === null) {
                    delete ctx.query[key];
                }
                else if (data !== ctx.query[key]) {
                    ctx.query[key] = '' + data;
                }
            }
            return ctx;
        });
        // Adjust origin param
        instance.on('beforeRequest', (ctx) => {
            const url = new URL(ctx.url);
            if (url.searchParams.has('origin')) {
                const origin = encodeURIComponent(url.searchParams.get('origin') || '').replace(/\./g, '%2E');
                ctx.query = {};
                url.searchParams.delete('origin');
                ctx.url = `${url}${url.search ? '&' : '?'}origin=${origin}`;
            }
            return ctx;
        });
        return instance;
    }
    /** Syntactic Sugar */
    // request handler
    get request() {
        return this.requestHandlerRef.value;
    }
    // userOptions
    get defaultOptions() {
        return this.defaultOptionsRef.value;
    }
    set defaultOptions(options) {
        this.defaultOptionsRef.value = options;
    }
    // defaultParams
    get defaultParams() {
        return this.defaultParamsRef.value;
    }
    set defaultParams(params) {
        this.defaultParamsRef.value = params;
    }
    /** Base methods encapsulation */
    get(query, options) {
        return this.request.get('', {
            query: query,
            ...options,
        });
    }
    post(data, options) {
        return this.request.post('', data, options);
    }
    async login(lgname, lgpassword, params, postOptions) {
        var _a, _b, _c, _d, _e, _f;
        this.defaultOptions.credentials = 'include';
        postOptions = postOptions || {};
        (_a = postOptions.retry) !== null && _a !== void 0 ? _a : (postOptions.retry = 3);
        if (postOptions.retry < 1) {
            throw new WikiSaikouError(WikiSaikouErrorCode.LOGIN_RETRY_LIMIT_EXCEEDED, 'The limit of the number of times to automatically re-login has been exceeded');
        }
        // FIXME: This is ugly
        let data;
        try {
            const res = await this.postWithToken('login', {
                action: 'login',
                lgname,
                lgpassword,
                ...params,
            }, { tokenName: 'lgtoken', ...postOptions });
            if ((_b = res === null || res === void 0 ? void 0 : res.data) === null || _b === void 0 ? void 0 : _b.login) {
                data = res.data;
            }
            else {
                throw res;
            }
        }
        catch (e) {
            if (e instanceof WikiSaikouError) {
                throw e;
            }
            else if ((e === null || e === void 0 ? void 0 : e.ok) === false) {
                return this.login(lgname, lgpassword, params, {
                    ...postOptions,
                    noCache: true,
                    retry: postOptions.retry - 1,
                });
            }
            else {
                throw new WikiSaikouError(WikiSaikouErrorCode.HTTP_ERROR, "The server returns an error, but it doesn't seem to be caused by MediaWiki", e);
            }
        }
        if (((_c = data === null || data === void 0 ? void 0 : data.login) === null || _c === void 0 ? void 0 : _c.result) !== 'Success') {
            throw new WikiSaikouError(WikiSaikouErrorCode.LOGIN_FAILED, ((_e = (_d = data === null || data === void 0 ? void 0 : data.login) === null || _d === void 0 ? void 0 : _d.reason) === null || _e === void 0 ? void 0 : _e.text) ||
                ((_f = data === null || data === void 0 ? void 0 : data.login) === null || _f === void 0 ? void 0 : _f.result) ||
                'Login failed with unknown reason', data);
        }
        return data.login;
    }
    async getUserInfo() {
        var _a;
        const { data } = await this.get({
            action: 'query',
            meta: 'userinfo',
            uiprop: ['groups', 'rights', 'blockinfo'],
        });
        return (_a = data === null || data === void 0 ? void 0 : data.query) === null || _a === void 0 ? void 0 : _a.userinfo;
    }
    /** Token Handler */
    async getTokens(type = ['csrf']) {
        this.defaultOptions.credentials = 'include';
        const { data } = await this.get({
            action: 'query',
            meta: 'tokens',
            type,
        });
        this.tokens = { ...this.tokens, ...data.query.tokens };
        return this.tokens;
    }
    async token(type = 'csrf', noCache = false) {
        if (!this.tokens[`${type}token`] || noCache) {
            delete this.tokens[`${type}token`];
            await this.getTokens([type]);
        }
        return this.tokens[`${type}token`];
    }
    async postWithToken(tokenType, body, options) {
        const { tokenName = 'token', retry = 3, noCache = false } = options || {};
        if (retry < 1) {
            throw new WikiSaikouError(WikiSaikouErrorCode.TOKEN_RETRY_LIMIT_EXCEEDED, 'The limit of the number of times to automatically re-acquire the token has been exceeded');
        }
        const token = await this.token(tokenType, noCache);
        const doRetry = () => this.postWithToken(tokenType, body, {
            tokenName,
            retry: retry - 1,
            noCache: true,
        });
        return this.post({
            [tokenName]: token,
            ...body,
        })
            .then((ctx) => {
            const data = ctx.data;
            if (MediaWikiApi.isBadTokenError(data)) {
                return doRetry();
            }
            return ctx;
        })
            .catch((err) => {
            const data = err.data;
            if (MediaWikiApi.isBadTokenError(data) || (err === null || err === void 0 ? void 0 : err.ok) === false) {
                return doRetry();
            }
            else if (typeof data === 'object' && data !== null) {
                return Promise.reject(data);
            }
            else {
                throw new WikiSaikouError(WikiSaikouErrorCode.HTTP_ERROR, 'The server returns an error, but it doesn’t seem to be caused by MediaWiki', err);
            }
        });
    }
    postWithEditToken(body) {
        return this.postWithToken('csrf', body);
    }
    static isBadTokenError(data) {
        var _a, _b, _c;
        return (((_a = data === null || data === void 0 ? void 0 : data.error) === null || _a === void 0 ? void 0 : _a.code) === 'badtoken' ||
            ((_b = data === null || data === void 0 ? void 0 : data.errors) === null || _b === void 0 ? void 0 : _b.some((i) => i.code === 'badtoken')) ||
            ['NeedToken', 'WrongToken'].includes((_c = data === null || data === void 0 ? void 0 : data.login) === null || _c === void 0 ? void 0 : _c.result));
    }
    async getMessages(ammessages, amlang = 'zh', options) {
        const { data } = await this.get({
            action: 'query',
            meta: 'allmessages',
            ammessages,
            amlang,
            ...options,
        });
        const result = {};
        data.query.allmessages.forEach(function (obj) {
            if (!obj.missing) {
                result[obj.name] = obj.content;
            }
        });
        return result;
    }
    async parseWikitext(wikitext, title, extraBody, options) {
        const { data } = await this.post({
            action: 'parse',
            title,
            text: wikitext,
            ...extraBody,
        }, options);
        return data.parse.text;
    }
}
exports.MediaWikiApi = MediaWikiApi;
exports.MwApi = MediaWikiApi;
class MediaWikiForeignApi extends MediaWikiApi {
    constructor(baseURL, options) {
        super(baseURL, {
            credentials: 'include',
            ...options,
        });
        this.defaultParams.origin = location.origin;
    }
}
exports.MediaWikiForeignApi = MediaWikiForeignApi;
exports.ForeignApi = MediaWikiForeignApi;
// Aliases
exports.default = MediaWikiApi;
// Errors
var WikiSaikouErrorCode;
(function (WikiSaikouErrorCode) {
    WikiSaikouErrorCode["HTTP_ERROR"] = "HTTP_ERROR";
    WikiSaikouErrorCode["LOGIN_FAILED"] = "LOGIN_FAILED";
    WikiSaikouErrorCode["LOGIN_RETRY_LIMIT_EXCEEDED"] = "LOGIN_RETRY_LIMIT_EXCEEDED";
    WikiSaikouErrorCode["TOKEN_RETRY_LIMIT_EXCEEDED"] = "TOKEN_RETRY_LIMIT_EXCEEDED";
})(WikiSaikouErrorCode || (exports.WikiSaikouErrorCode = WikiSaikouErrorCode = {}));
class WikiSaikouError extends Error {
    constructor(code, message = '', cause) {
        super();
        this.code = code;
        this.message = message;
        this.cause = cause;
        this.name = 'WikiSaikouError';
    }
}
exports.WikiSaikouError = WikiSaikouError;
//# sourceMappingURL=index.js.map